@@ -53,6 +53,8 @@ import java.security.SecureRandom;
 import java.util.ArrayList;
 import java.util.Collections;
 import java.util.List;
+import java.util.concurrent.CountDownLatch;
+import java.util.concurrent.TimeUnit;
 
 /**
  * Convenience class used to build an Undertow server.
@@ -112,7 +114,7 @@ public final class Undertow {
     }
 
     public synchronized void start() {
-        UndertowLogger.ROOT_LOGGER.debugf("starting undertow server %s", this);
+        UndertowLogger.ROOT_LOGGER.infof("starting server: %s", Version.getFullVersionString());
         xnio = Xnio.getInstance(Undertow.class.getClassLoader());
         channels = new ArrayList<>();
         try {
@@ -154,6 +156,7 @@ public final class Undertow {
             for (ListenerConfig listener : listeners) {
                 UndertowLogger.ROOT_LOGGER.debugf("Configuring listener with protocol %s for interface %s and port %s", listener.type, listener.host, listener.port);
                 final HttpHandler rootHandler = listener.rootHandler != null ? listener.rootHandler : this.rootHandler;
+                OptionMap socketOptionsWithOverrides = OptionMap.builder().addAll(socketOptions).addAll(listener.overrideSocketOptions).getMap();
                 if (listener.type == ListenerType.AJP) {
                     AjpOpenListener openListener = new AjpOpenListener(buffers, serverOptions);
                     openListener.setRootHandler(rootHandler);
@@ -165,7 +168,6 @@ public final class Undertow {
                         finalListener = openListener;
                     }
                     ChannelListener<AcceptingChannel<StreamConnection>> acceptListener = ChannelListeners.openListenerAdapter(finalListener);
-                    OptionMap socketOptionsWithOverrides = OptionMap.builder().addAll(socketOptions).addAll(listener.overrideSocketOptions).getMap();
                     AcceptingChannel<? extends StreamConnection> server = worker.createStreamConnectionServer(new InetSocketAddress(Inet4Address.getByName(listener.host), listener.port), acceptListener, socketOptionsWithOverrides);
                     server.resumeAccepts();
                     channels.add(server);
@@ -188,7 +190,6 @@ public final class Undertow {
                         }
 
                         ChannelListener<AcceptingChannel<StreamConnection>> acceptListener = ChannelListeners.openListenerAdapter(finalListener);
-                        OptionMap socketOptionsWithOverrides = OptionMap.builder().addAll(socketOptions).addAll(listener.overrideSocketOptions).getMap();
                         AcceptingChannel<? extends StreamConnection> server = worker.createStreamConnectionServer(new InetSocketAddress(Inet4Address.getByName(listener.host), listener.port), acceptListener, socketOptionsWithOverrides);
                         server.resumeAccepts();
                         channels.add(server);
@@ -201,12 +202,10 @@ public final class Undertow {
 
                         if (http2) {
                             AlpnOpenListener alpn = new AlpnOpenListener(buffers, undertowOptions, httpOpenListener);
-                            if (http2) {
-                                Http2OpenListener http2Listener = new Http2OpenListener(buffers, undertowOptions);
-                                http2Listener.setRootHandler(rootHandler);
-                                alpn.addProtocol(Http2OpenListener.HTTP2, http2Listener, 10);
-                                alpn.addProtocol(Http2OpenListener.HTTP2_14, http2Listener, 7);
-                            }
+                            Http2OpenListener http2Listener = new Http2OpenListener(buffers, undertowOptions);
+                            http2Listener.setRootHandler(rootHandler);
+                            alpn.addProtocol(Http2OpenListener.HTTP2, http2Listener, 10);
+                            alpn.addProtocol(Http2OpenListener.HTTP2_14, http2Listener, 7);
                             openListener = alpn;
                         } else {
                             openListener = httpOpenListener;
@@ -216,15 +215,14 @@ public final class Undertow {
                         if (listener.sslContext != null) {
                             xnioSsl = new UndertowXnioSsl(xnio, OptionMap.create(Options.USE_DIRECT_BUFFERS, true), listener.sslContext);
                         } else {
-                            OptionMap.Builder builder = OptionMap.builder();
-                            builder.addAll(listener.overrideSocketOptions);
-                            if (!listener.overrideSocketOptions.contains(Options.SSL_PROTOCOL)) {
+                            OptionMap.Builder builder = OptionMap.builder()
+                                    .addAll(socketOptionsWithOverrides);
+                            if (!socketOptionsWithOverrides.contains(Options.SSL_PROTOCOL)) {
                                 builder.set(Options.SSL_PROTOCOL, "TLSv1.2");
                             }
                             xnioSsl = new UndertowXnioSsl(xnio, OptionMap.create(Options.USE_DIRECT_BUFFERS, true), JsseSslUtils.createSSLContext(listener.keyManagers, listener.trustManagers, new SecureRandom(), builder.getMap()));
                         }
 
-                        OptionMap socketOptionsWithOverrides = OptionMap.builder().addAll(socketOptions).addAll(listener.overrideSocketOptions).getMap();
                         AcceptingChannel<? extends StreamConnection> sslServer;
                         if (listener.useProxyProtocol) {
                             ChannelListener<AcceptingChannel<StreamConnection>> acceptListener = ChannelListeners.openListenerAdapter(new ProxyProtocolOpenListener(openListener, xnioSsl, buffers, socketOptionsWithOverrides));
@@ -251,7 +249,7 @@ public final class Undertow {
     }
 
     public synchronized void stop() {
-        UndertowLogger.ROOT_LOGGER.debugf("stopping undertow server %s", this);
+        UndertowLogger.ROOT_LOGGER.infof("stopping server: %s", Version.getFullVersionString());
         if (channels != null) {
             for (AcceptingChannel<? extends StreamConnection> channel : channels) {
                 IoUtils.safeClose(channel);
@@ -263,7 +261,20 @@ public final class Undertow {
          * Only shutdown the worker if it was created during start()
          */
         if (internalWorker && worker != null) {
-            worker.shutdownNow();
+            Integer shutdownTimeoutMillis = serverOptions.get(UndertowOptions.SHUTDOWN_TIMEOUT);
+            worker.shutdown();
+            try {
+                if (shutdownTimeoutMillis == null) {
+                    worker.awaitTermination();
+                } else {
+                    if (!worker.awaitTermination(shutdownTimeoutMillis, TimeUnit.MILLISECONDS)) {
+                        worker.shutdownNow();
+                    }
+                }
+            } catch (InterruptedException e) {
+                worker.shutdownNow();
+                throw new RuntimeException(e);
+            }
             worker = null;
         }
         xnio = null;
@@ -547,7 +558,7 @@ public final class Undertow {
          * to the various worker-related configuration (ioThreads, workerThreads, workerOptions)
          * when {@link Undertow#start()} is called.
          * Additionally, this newly created worker will be shutdown when {@link Undertow#stop()} is called.
-         * <br/>
+         * <br>
          * <p>
          * When non-null, the provided {@link XnioWorker} will be reused instead of creating a new {@link XnioWorker}
          * when {@link Undertow#start()} is called.
@@ -572,6 +583,7 @@ public final class Undertow {
         private final OpenListener openListener;
         private final UndertowXnioSsl ssl;
         private final AcceptingChannel<? extends StreamConnection> channel;
+        private volatile boolean suspended = false;
 
         public ListenerInfo(String protcol, SocketAddress address, OpenListener openListener, UndertowXnioSsl ssl, AcceptingChannel<? extends StreamConnection> channel) {
             this.protcol = protcol;
@@ -603,6 +615,37 @@ public final class Undertow {
             }
         }
 
+        public synchronized void suspend() {
+            suspended = true;
+            channel.suspendAccepts();
+            CountDownLatch latch = new CountDownLatch(1);
+            //the channel may be in the middle of an accept, we need to close from the IO thread
+            channel.getIoThread().execute(new Runnable() {
+                @Override
+                public void run() {
+                    try {
+                        openListener.closeConnections();
+                    } finally {
+                        latch.countDown();
+                    }
+                }
+            });
+            try {
+                latch.await();
+            } catch (InterruptedException e) {
+                throw new RuntimeException(e);
+            }
+        }
+
+        public synchronized void resume() {
+            suspended = false;
+            channel.resumeAccepts();
+        }
+
+        public boolean isSuspended() {
+            return suspended;
+        }
+
         public ConnectorStatistics getConnectorStatistics() {
             return openListener.getConnectorStatistics();
         }
