@@ -68,6 +68,7 @@ public interface UndertowLogger extends BasicLogger {
      * attacker to fill up the logs by intentionally causing IO exceptions.
      */
     UndertowLogger REQUEST_IO_LOGGER = Logger.getMessageLogger(UndertowLogger.class, UndertowLogger.class.getPackage().getName() + ".request.io");
+    UndertowLogger ERROR_RESPONSE = Logger.getMessageLogger(UndertowLogger.class, UndertowLogger.class.getPackage().getName() + ".request.error-response");
 
     @LogMessage(level = ERROR)
     @Message(id = 5001, value = "An exception occurred processing the request")
@@ -400,9 +401,9 @@ public interface UndertowLogger extends BasicLogger {
     @Message(id = 5086, value = "Failed to accept SSL request")
     void failedToAcceptSSLRequest(@Cause Exception e);
 
-    @LogMessage(level = ERROR)
-    @Message(id = 5087, value = "Failed to use the server order")
-    void failedToUseServerOrder(@Cause ReflectiveOperationException e);
+//    @LogMessage(level = ERROR)
+//    @Message(id = 5087, value = "Failed to use the server order")
+//    void failedToUseServerOrder(@Cause ReflectiveOperationException e);
 
     @LogMessage(level = ERROR)
     @Message(id = 5088, value = "Failed to execute ServletOutputStream.closeAsync() on IO thread")
