@@ -92,13 +92,17 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
     public final ConduitListener<StreamSinkConduit> requestFinishListener = new ConduitListener<StreamSinkConduit>() {
         @Override
         public void handleEvent(StreamSinkConduit channel) {
-            currentRequest.terminateRequest();
+            if(currentRequest != null) {
+                currentRequest.terminateRequest();
+            }
         }
     };
     public final ConduitListener<StreamSourceConduit> responseFinishedListener = new ConduitListener<StreamSourceConduit>() {
         @Override
         public void handleEvent(StreamSourceConduit channel) {
-            currentRequest.terminateResponse();
+            if(currentRequest != null) {
+                currentRequest.terminateResponse();
+            }
         }
     };
 
@@ -437,8 +441,10 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
     }
     private void handleError(IOException exception) {
         UndertowLogger.REQUEST_IO_LOGGER.ioException(exception);
-        safeClose(connection);
         currentRequest.setFailed(exception);
+        currentRequest = null;
+        pendingResponse = null;
+        safeClose(connection);
     }
 
     public StreamConnection performUpgrade() throws IOException {
@@ -550,8 +556,10 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
                             UndertowLogger.CLIENT_LOGGER.debugf(e, "Connection closed with IOException");
                         }
                         try {
-                            currentRequest.setFailed(e);
-                            currentRequest = null;
+                            if (currentRequest != null) {
+                                currentRequest.setFailed(e);
+                                currentRequest = null;
+                            }
                             pendingResponse = null;
                         } finally {
                             safeClose(channel, HttpClientConnection.this);
@@ -569,8 +577,10 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
                         channel.suspendReads();
                         try {
                             // Cancel the current active request
-                            currentRequest.setFailed(new IOException(MESSAGES.connectionClosed()));
-                            currentRequest = null;
+                            if (currentRequest != null) {
+                                currentRequest.setFailed(new IOException(MESSAGES.connectionClosed()));
+                                currentRequest = null;
+                            }
                             pendingResponse = null;
                         } finally {
                             safeClose(HttpClientConnection.this);
@@ -647,7 +657,10 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
                                 sinkChannel.setWriteListener(ChannelListeners.flushingChannelListener(null, null));
                                 sinkChannel.resumeWrites();
                             }
-                            currentRequest.terminateRequest();
+                            if(currentRequest != null) {
+                                //we need the null check as flushing the response may have terminated the request
+                                currentRequest.terminateRequest();
+                            }
                         }
                     }
                 }
