@@ -29,10 +29,15 @@ import static io.undertow.util.Headers.TRANSFER_ENCODING;
 import java.io.IOException;
 import java.net.SocketAddress;
 import java.nio.channels.ClosedChannelException;
+import java.util.Arrays;
+import java.util.HashMap;
 import java.util.List;
 import java.util.Map;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.CopyOnWriteArrayList;
+import java.util.concurrent.TimeUnit;
+import java.util.concurrent.atomic.AtomicLong;
+import java.util.function.Supplier;
 
 import io.undertow.client.ClientStatistics;
 import io.undertow.protocols.http2.Http2DataStreamSinkChannel;
@@ -80,11 +85,17 @@ public class Http2ClientConnection implements ClientConnection {
 
     private final Map<Integer, Http2ClientExchange> currentExchanges = new ConcurrentHashMap<>();
 
+    private static final AtomicLong PING_COUNTER = new AtomicLong();
+
+
     private boolean initialUpgradeRequest;
     private final String defaultHost;
     private final ClientStatistics clientStatistics;
     private final List<ChannelListener<ClientConnection>> closeListeners = new CopyOnWriteArrayList<>();
     private final boolean secure;
+
+    private final Map<PingKey, PingListener> outstandingPings = new HashMap<>();
+
     private final ChannelListener<Http2Channel> closeTask = new ChannelListener<Http2Channel>() {
         @Override
         public void handleEvent(Http2Channel channel) {
@@ -176,13 +187,27 @@ public class Http2ClientConnection implements ClientConnection {
             clientCallback.failed(e);
             return;
         }
-        final Http2ClientExchange exchange = new Http2ClientExchange(this, sinkChannel, request);
+        Http2ClientExchange exchange = new Http2ClientExchange(this, sinkChannel, request);
         currentExchanges.put(sinkChannel.getStreamId(), exchange);
 
         sinkChannel.setTrailersProducer(new Http2DataStreamSinkChannel.TrailersProducer() {
             @Override
             public HeaderMap getTrailers() {
-                return exchange.getAttachment(HttpAttachments.RESPONSE_TRAILERS);
+                HeaderMap attachment = exchange.getAttachment(HttpAttachments.RESPONSE_TRAILERS);
+                Supplier<HeaderMap> supplier = exchange.getAttachment(HttpAttachments.RESPONSE_TRAILER_SUPPLIER);
+                if(attachment != null && supplier == null) {
+                    return attachment;
+                } else if(attachment == null && supplier != null) {
+                    return supplier.get();
+                } else if(attachment != null) {
+                    HeaderMap supplied = supplier.get();
+                    for(HeaderValues k : supplied) {
+                        attachment.putAll(k.getHeaderName(), k);
+                    }
+                    return attachment;
+                } else {
+                    return null;
+                }
             }
         });
 
@@ -329,6 +354,36 @@ public class Http2ClientConnection implements ClientConnection {
         closeListeners.add(listener);
     }
 
+    @Override
+    public boolean isPingSupported() {
+        return true;
+    }
+
+    @Override
+    public void sendPing(PingListener listener, long timeout, TimeUnit timeUnit) {
+        long count = PING_COUNTER.incrementAndGet();
+        byte[] data = new byte[8];
+        data[0] = (byte) count;
+        data[1] = (byte)(count << 8);
+        data[2] = (byte)(count << 16);
+        data[3] = (byte)(count << 24);
+        data[4] = (byte)(count << 32);
+        data[5] = (byte)(count << 40);
+        data[6] = (byte)(count << 48);
+        data[7] = (byte)(count << 54);
+        final PingKey key = new PingKey(data);
+        outstandingPings.put(key, listener);
+        if(timeout > 0) {
+            http2Channel.getIoThread().executeAfter(() -> {
+                PingListener listener1 = outstandingPings.remove(key);
+                if(listener1 != null) {
+                    listener1.failed(UndertowMessages.MESSAGES.pingTimeout());
+                }
+            }, timeout, timeUnit);
+        }
+        http2Channel.sendPing(data, (channel, exception) -> listener.failed(exception));
+    }
+
     private class Http2ReceiveListener implements ChannelListener<Http2Channel> {
 
         @Override
@@ -339,7 +394,7 @@ public class Http2ClientConnection implements ClientConnection {
                     final Http2StreamSourceChannel streamSourceChannel = (Http2StreamSourceChannel) result;
 
                     int statusCode = Integer.parseInt(streamSourceChannel.getHeaders().getFirst(STATUS));
-                    final Http2ClientExchange request = currentExchanges.get(streamSourceChannel.getStreamId());
+                    Http2ClientExchange request = currentExchanges.get(streamSourceChannel.getStreamId());
                     if(statusCode < 200) {
                         //this is an informational response 1xx response
                         if(statusCode == 100) {
@@ -444,8 +499,36 @@ public class Http2ClientConnection implements ClientConnection {
             if (!frame.isAck()) {
                 //server side ping, return it
                 frame.getHttp2Channel().sendPing(id);
+            } else {
+                PingListener listener = outstandingPings.remove(new PingKey(id));
+                if(listener != null) {
+                    listener.acknowledged();
+                }
             }
         }
 
     }
+
+    private static final class PingKey{
+        private final byte[] data;
+
+        private PingKey(byte[] data) {
+            this.data = data;
+        }
+
+        @Override
+        public boolean equals(Object o) {
+            if (this == o) return true;
+            if (o == null || getClass() != o.getClass()) return false;
+
+            PingKey pingKey = (PingKey) o;
+
+            return Arrays.equals(data, pingKey.data);
+        }
+
+        @Override
+        public int hashCode() {
+            return Arrays.hashCode(data);
+        }
+    }
 }
