@@ -486,7 +486,7 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
                 boolean ack = Bits.anyAreSet(frameParser.flags, PING_FLAG_ACK);
                 channel = new Http2PingStreamSourceChannel(this, pingParser.getData(), ack);
                 if(!ack) { //not an ack from one of our pings, so send it back
-                    sendPing(pingParser.getData(), null, true);
+                    sendPing(pingParser.getData(),  new Http2ControlMessageExceptionHandler(), true);
                 }
                 break;
             }
@@ -587,17 +587,22 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
     }
 
     protected void lastDataRead() {
-        lastDataRead = true;
-        if(!peerGoneAway) {
-            //we just close the connection, as the peer has performed an unclean close
-            IoUtils.safeClose(this);
-        } else {
-            peerGoneAway = true;
-            if(!thisGoneAway) {
-                //we send a goaway message, and then close
-                sendGoAway(ERROR_CONNECT_ERROR);
+        final boolean peerGoneAway;
+        synchronized (this) {
+            lastDataRead = true;
+            peerGoneAway = this.peerGoneAway;
+            if(peerGoneAway) {
+                if(!thisGoneAway) {
+                    //we send a goaway message, and then close
+                    sendGoAway(ERROR_CONNECT_ERROR);
+                }
             }
         }
+        if (!peerGoneAway) {
+            //we just close the connection, as the peer has performed an unclean close
+            IoUtils.safeClose(this);
+        }
+
     }
 
     @Override
