@@ -101,6 +101,18 @@ public class Connectors {
         }
     }
 
+    /**
+     * Adds the cookie into the response header map. This should be called
+     * before the response is started.
+     *
+     * @param exchange The server exchange
+     * @param cookie   The cookie
+     */
+    public static void addCookie(final HttpServerExchange exchange, Cookie cookie) {
+        boolean enableRfc6265Validation = exchange.getConnection().getUndertowOptions().get(UndertowOptions.ENABLE_RFC6265_COOKIE_VALIDATION, UndertowOptions.DEFAULT_ENABLE_RFC6265_COOKIE_VALIDATION);
+        exchange.getResponseHeaders().add(Headers.SET_COOKIE, getCookieString(cookie, enableRfc6265Validation));
+    }
+
     /**
      * Attached buffered data to the exchange. The will generally be used to allow data to be re-read.
      *
@@ -119,20 +131,24 @@ public class Connectors {
             System.arraycopy(buffers, 0, newArray, existing.length, buffers.length);
         }
         exchange.putAttachment(HttpServerExchange.BUFFERED_REQUEST_DATA, newArray); //todo: force some kind of wakeup?
-        exchange.addExchangeCompleteListener(new ExchangeCompletionListener() {
-            @Override
-            public void exchangeEvent(HttpServerExchange exchange, NextListener nextListener) {
-                PooledByteBuffer[] bufs = exchange.getAttachment(HttpServerExchange.BUFFERED_REQUEST_DATA);
-                if (bufs != null) {
-                    for (PooledByteBuffer i : bufs) {
-                        if(i != null) {
-                            i.close();
-                        }
+        exchange.addExchangeCompleteListener(BufferedRequestDataCleanupListener.INSTANCE);
+    }
+
+    private enum BufferedRequestDataCleanupListener implements ExchangeCompletionListener {
+        INSTANCE;
+
+        @Override
+        public void exchangeEvent(HttpServerExchange exchange, NextListener nextListener) {
+            PooledByteBuffer[] bufs = exchange.getAttachment(HttpServerExchange.BUFFERED_REQUEST_DATA);
+            if (bufs != null) {
+                for (PooledByteBuffer i : bufs) {
+                    if(i != null) {
+                        i.close();
                     }
                 }
-                nextListener.proceed();
             }
-        });
+            nextListener.proceed();
+        }
     }
 
     public static void terminateRequest(final HttpServerExchange exchange) {
@@ -223,6 +239,14 @@ public class Connectors {
             header.append("; Comment=");
             header.append(cookie.getComment());
         }
+        if (cookie.isSameSite()) {
+            if (cookie.getSameSiteMode() != null && !cookie.getSameSiteMode().isEmpty()) {
+                header.append("; SameSite=");
+                header.append(cookie.getSameSiteMode());
+            } else {
+                header.append("; SameSite");
+            }
+        }
         return header.toString();
     }
 
@@ -267,6 +291,14 @@ public class Connectors {
                 header.append(DateUtils.toOldCookieDateString(expires));
             }
         }
+        if (cookie.isSameSite()) {
+            if (cookie.getSameSiteMode() != null && !cookie.getSameSiteMode().isEmpty()) {
+                header.append("; SameSite=");
+                header.append(cookie.getSameSiteMode());
+            } else {
+                header.append("; SameSite");
+            }
+        }
         return header.toString();
 
     }
@@ -327,6 +359,14 @@ public class Connectors {
             header.append("; Comment=");
             LegacyCookieSupport.maybeQuote(header, cookie.getComment());
         }
+        if (cookie.isSameSite()) {
+            if (cookie.getSameSiteMode() != null && !cookie.getSameSiteMode().isEmpty()) {
+                header.append("; SameSite=");
+                header.append(cookie.getSameSiteMode());
+            } else {
+                header.append("; SameSite");
+            }
+        }
         return header.toString();
     }
 
@@ -335,7 +375,7 @@ public class Connectors {
             exchange.setInCall(true);
             handler.handleRequest(exchange);
             exchange.setInCall(false);
-            boolean resumed = exchange.runResumeReadWrite();
+            boolean resumed = exchange.isResumed();
             if (exchange.isDispatched()) {
                 if (resumed) {
                     UndertowLogger.REQUEST_LOGGER.resumedAndDispatched();
@@ -353,12 +393,14 @@ public class Connectors {
                         executor.execute(dispatchTask);
                     } catch (RejectedExecutionException e) {
                         UndertowLogger.REQUEST_LOGGER.debug("Failed to dispatch to worker", e);
-                        exchange.setStatusCode(StatusCodes.INTERNAL_SERVER_ERROR);
+                        exchange.setStatusCode(StatusCodes.SERVICE_UNAVAILABLE);
                         exchange.endExchange();
                     }
                 }
             } else if (!resumed) {
                 exchange.endExchange();
+            } else {
+                exchange.runResumeReadWrite();
             }
         } catch (Throwable t) {
             exchange.putAttachment(DefaultResponseListener.EXCEPTION, t);
@@ -441,7 +483,7 @@ public class Connectors {
                 URLUtils.parsePathParams(encodedPath.substring(i + 1), exchange, charset, decode, maxParameters);
                 return;
             } else if(c == '%' || c == '+') {
-                requiresDecode = true;
+                requiresDecode = decode;
             }
         }
 
