@@ -55,6 +55,7 @@ import org.xnio.channels.ConnectedChannel;
 import org.xnio.channels.StreamSinkChannel;
 import org.xnio.channels.StreamSourceChannel;
 import org.xnio.channels.SuspendableWriteChannel;
+
 import io.undertow.UndertowLogger;
 import io.undertow.UndertowMessages;
 import io.undertow.UndertowOptions;
@@ -110,7 +111,10 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
     private volatile long frameDataRemaining;
     private volatile R receiver;
 
-    private volatile boolean receivesSuspended = true;
+    private volatile boolean receivesSuspendedByUser = true;
+    private volatile boolean receivesSuspendedTooManyQueuedMessages = false;
+    private volatile boolean receivesSuspendedTooManyBuffers = false;
+
 
     @SuppressWarnings("unused")
     private volatile int readsBroken = 0;
@@ -125,11 +129,22 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
     private final List<ChannelListener<C>> closeTasks = new CopyOnWriteArrayList<>();
     private volatile boolean flushingSenders = false;
 
+    private boolean partialRead = false;
+
     @SuppressWarnings("unused")
     private volatile int outstandingBuffers;
-    private volatile AtomicIntegerFieldUpdater<AbstractFramedChannel> outstandingBuffersUpdater = AtomicIntegerFieldUpdater.newUpdater(AbstractFramedChannel.class, "outstandingBuffers");
+    private static final AtomicIntegerFieldUpdater<AbstractFramedChannel> outstandingBuffersUpdater = AtomicIntegerFieldUpdater.newUpdater(AbstractFramedChannel.class, "outstandingBuffers");
 
     private final LinkedBlockingDeque<Runnable> taskRunQueue = new LinkedBlockingDeque<>();
+    private final Runnable taskRunQueueRunnable = new Runnable() {
+        @Override
+        public void run() {
+            Runnable runnable;
+            while ((runnable = taskRunQueue.poll()) != null) {
+                runnable.run();
+            }
+        }
+    };
     private final OptionMap settings;
 
     /**
@@ -138,19 +153,29 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
     private volatile boolean requireExplicitFlush = false;
     private volatile boolean readChannelDone = false;
 
+    private final int queuedFrameHighWaterMark;
+    private final int queuedFrameLowWaterMark;
+
     private final ReferenceCountedPooled.FreeNotifier freeNotifier = new ReferenceCountedPooled.FreeNotifier() {
         @Override
         public void freed() {
             int res = outstandingBuffersUpdater.decrementAndGet(AbstractFramedChannel.this);
-            if(!receivesSuspended && res == maxQueuedBuffers - 1) {
-                synchronized (AbstractFramedChannel.this) {
-                    if(outstandingBuffersUpdater.get(AbstractFramedChannel.this) < maxQueuedBuffers) {
-                        if(UndertowLogger.REQUEST_IO_LOGGER.isTraceEnabled()) {
-                            UndertowLogger.REQUEST_IO_LOGGER.tracef("Resuming reads on %s as buffers have been consumed", AbstractFramedChannel.this);
+            if (!receivesSuspendedByUser && res == maxQueuedBuffers - 1) {
+                //we need to do the resume in the IO thread, as there is a risk of deadlock otherwise, as the calling thread is an application thread
+                //and may hold a lock on a stream source channel, see UNDERTOW-1312
+                getIoThread().execute(new Runnable() {
+                    @Override
+                    public void run() {
+                        synchronized (AbstractFramedChannel.this) {
+                            if (outstandingBuffersUpdater.get(AbstractFramedChannel.this) < maxQueuedBuffers) {
+                                if (UndertowLogger.REQUEST_IO_LOGGER.isTraceEnabled()) {
+                                    UndertowLogger.REQUEST_IO_LOGGER.tracef("Resuming reads on %s as buffers have been consumed", AbstractFramedChannel.this);
+                                }
+                                new UpdateResumeState(null, false, null).run();
+                            }
                         }
-                        channel.getSourceChannel().resumeReads();
                     }
-                }
+                });
             }
         }
     };
@@ -214,6 +239,8 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
         FrameCloseListener closeListener = new FrameCloseListener();
         connectedStreamChannel.getSinkChannel().getCloseSetter().set(closeListener);
         connectedStreamChannel.getSourceChannel().getCloseSetter().set(closeListener);
+        this.queuedFrameHighWaterMark = settings.get(UndertowOptions.QUEUED_FRAMES_HIGH_WATER_MARK, 50);
+        this.queuedFrameLowWaterMark = settings.get(UndertowOptions.QUEUED_FRAMES_LOW_WATER_MARK, 10);
     }
 
     protected IdleTimeoutConduit createIdleTimeoutChannel(StreamConnection connectedStreamChannel) {
@@ -223,19 +250,10 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
     void runInIoThread(Runnable task) {
         this.taskRunQueue.add(task);
         try {
-            getIoThread().execute(new Runnable() {
-                @Override
-                public void run() {
-                    while (!taskRunQueue.isEmpty()) {
-                        taskRunQueue.poll().run();
-                    }
-                }
-            });
+            getIoThread().execute(taskRunQueueRunnable);
         } catch (RejectedExecutionException e) {
             //thread is shutting down
-            while (!taskRunQueue.isEmpty()) {
-                taskRunQueue.poll().run();
-            }
+            ShutdownFallbackExecutor.execute(taskRunQueueRunnable);
         }
     }
 
@@ -324,166 +342,185 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
      * existing source channels. In general if you suspend receives or don't have some other way
      * of calling this method then it can prevent frame channels for being fully consumed.
      */
-    public synchronized R receive() throws IOException {
-        if (readChannelDone && receiver == null) {
-            //we have received the last frame, we just shut down and return
-            //it would probably make more sense to have the last channel responsible for this
-            //however it is much simpler just to have it here
-            if(readData != null) {
-                readData.close();
-                readData = null;
-            }
-            channel.getSourceChannel().suspendReads();
-            channel.getSourceChannel().shutdownReads();
-            return null;
-        }
-        ReferenceCountedPooled pooled = this.readData;
-        boolean hasData;
-        if (pooled == null) {
-            pooled = allocateReferenceCountedBuffer();
-            if (pooled == null) {
-                return null;
-            }
-            hasData = false;
-        } else if(pooled.isFreed()) {
-            //we attempt to re-used an existing buffer
-            if(!pooled.tryUnfree()) {
-                pooled = allocateReferenceCountedBuffer();
-                if (pooled == null) {
-                    return null;
-                }
-            } else {
-                pooled.getBuffer().limit(pooled.getBuffer().capacity());
-            }
-            hasData = false;
-        } else {
-            hasData = pooled.getBuffer().hasRemaining();
-        }
-        boolean forceFree = false;
-        int read = 0;
+    public R receive() throws IOException {
+        // store in a local variable to prevent invoking lastDataRead twice
+        boolean receivedMinusOne = false;
         try {
-            if (!hasData) {
-                pooled.getBuffer().clear();
-                read = channel.getSourceChannel().read(pooled.getBuffer());
-                if (read == 0) {
-                    //no data, we just free the buffer
-                    forceFree = true;
-                    return null;
-                } else if (read == -1) {
-                    forceFree = true;
-                    readChannelDone = true;
-                    lastDataRead();
-                    return null;
-                } else if(isLastFrameReceived() && frameDataRemaining == 0) {
-                    //we got data, although we should have received the last frame
-                    forceFree = true;
-                    markReadsBroken(new ClosedChannelException());
-                }
-                pooled.getBuffer().flip();
-            }
-            if (frameDataRemaining > 0) {
-                if (frameDataRemaining >= pooled.getBuffer().remaining()) {
-                    frameDataRemaining -= pooled.getBuffer().remaining();
-                    if(receiver != null) {
-                        //we still create a pooled view, this means that if the buffer is still active we can re-used it
-                        //which prevents attacks based on sending lots of small fragments
-                        ByteBuffer buf = pooled.getBuffer().duplicate();
-                        pooled.getBuffer().position(pooled.getBuffer().limit());
-                        PooledByteBuffer frameData = pooled.createView(buf);
-                        receiver.dataReady(null, frameData);
-                    } else {
-                        //we are dropping a frame
-                        pooled.close();
+            synchronized (this) {
+                if (readChannelDone && receiver == null) {
+                    //we have received the last frame, we just shut down and return
+                    //it would probably make more sense to have the last channel responsible for this
+                    //however it is much simpler just to have it here
+                    if(readData != null) {
+                        readData.close();
                         readData = null;
                     }
-                    if(frameDataRemaining == 0) {
-                        receiver = null;
-                    }
+                    channel.getSourceChannel().suspendReads();
+                    channel.getSourceChannel().shutdownReads();
                     return null;
-                } else {
-                    ByteBuffer buf = pooled.getBuffer().duplicate();
-                    buf.limit((int) (buf.position() + frameDataRemaining));
-                    pooled.getBuffer().position((int) (pooled.getBuffer().position() + frameDataRemaining));
-                    frameDataRemaining = 0;
-                    PooledByteBuffer frameData = pooled.createView(buf);
-                    if(receiver != null) {
-                        receiver.dataReady(null, frameData);
-                    } else{
-                        //we are dropping the frame
-                        frameData.close();
-                    }
-                    receiver = null;
                 }
-                //if we read data into a frame we just return immediately, even if there is more remaining
-                //see https://issues.jboss.org/browse/UNDERTOW-410
-                //basically if we don't do this we loose some message ordering semantics
-                //as the second message may be processed before the first one
-
-                //this is problematic for HTTPS, where the read listener may also be invoked by a queued task
-                //and not by the selector mechanism
-                return null;
-            }
-            FrameHeaderData data = parseFrame(pooled.getBuffer());
-            if (data != null) {
-                PooledByteBuffer frameData;
-                if (data.getFrameLength() >= pooled.getBuffer().remaining()) {
-                    frameDataRemaining = data.getFrameLength() - pooled.getBuffer().remaining();
-                    frameData = pooled.createView(pooled.getBuffer().duplicate());
-                    pooled.getBuffer().position(pooled.getBuffer().limit());
-                } else {
-                    ByteBuffer buf = pooled.getBuffer().duplicate();
-                    buf.limit((int) (buf.position() + data.getFrameLength()));
-                    pooled.getBuffer().position((int) (pooled.getBuffer().position() + data.getFrameLength()));
-                    frameData = pooled.createView(buf);
-                }
-                AbstractFramedStreamSourceChannel<?, ?, ?> existing = data.getExistingChannel();
-                if (existing != null) {
-                    if (data.getFrameLength() > frameData.getBuffer().remaining()) {
-                        receiver = (R) existing;
+                partialRead = false;
+                boolean requiresReinvoke = false;
+                int reinvokeDataRemaining = 0;
+                ReferenceCountedPooled pooled = this.readData;
+                boolean hasData = false;
+                if (pooled == null) {
+                    pooled = allocateReferenceCountedBuffer();
+                    if (pooled == null) {
+                        return null;
                     }
-                    existing.dataReady(data, frameData);
-                    if(isLastFrameReceived()) {
-                        handleLastFrame(existing);
-                    }
-                    return null;
-                } else {
-                    boolean moreData = data.getFrameLength() > frameData.getBuffer().remaining();
-                    R newChannel = createChannel(data, frameData);
-                    if (newChannel != null) {
-                        if (moreData) {
-                            receiver = newChannel;
+                } else if(pooled.isFreed()) {
+                    //we attempt to re-used an existing buffer
+                    if(!pooled.tryUnfree()) {
+                        pooled = allocateReferenceCountedBuffer();
+                        if (pooled == null) {
+                            return null;
                         }
+                    }
+                    pooled.getBuffer().clear();
+                } else {
+                    hasData = pooled.getBuffer().hasRemaining();
+                    pooled.getBuffer().compact();
+                }
+                boolean forceFree = false;
+                int read = 0;
+                try {
+                    read = channel.getSourceChannel().read(pooled.getBuffer());
+                    if (read == 0 && !hasData) {
+                        //no data, we just free the buffer
+                        forceFree = true;
+                        return null;
+                    } else if (read == -1 && !hasData) {
+                        forceFree = true;
+                        receivedMinusOne = readChannelDone = true;
+                        return null;
+                    } else if(isLastFrameReceived() && frameDataRemaining == 0) {
+                        //we got data, although we should have received the last frame
+                        forceFree = true;
+                        markReadsBroken(new ClosedChannelException());
+                    }
+                    pooled.getBuffer().flip();
+                    if(read == -1) {
+                        requiresReinvoke = true;
+                        reinvokeDataRemaining = pooled.getBuffer().remaining();
+                    }
+                    if (frameDataRemaining > 0) {
+                        if (frameDataRemaining >= pooled.getBuffer().remaining()) {
+                            frameDataRemaining -= pooled.getBuffer().remaining();
+                            if(receiver != null) {
+                                //we still create a pooled view, this means that if the buffer is still active we can re-used it
+                                //which prevents attacks based on sending lots of small fragments
+                                PooledByteBuffer frameData = pooled.createView();
+                                receiver.dataReady(null, frameData);
+                            } else {
+                                //we are dropping a frame
+                                pooled.close();
+                                readData = null;
+                            }
+                            if(frameDataRemaining == 0) {
+                                receiver = null;
+                            }
+                            return null;
+                        } else {
+                            PooledByteBuffer frameData = pooled.createView((int) frameDataRemaining);
+                            frameDataRemaining = 0;
+                            if(receiver != null) {
+                                receiver.dataReady(null, frameData);
+                            } else{
+                                //we are dropping the frame
+                                frameData.close();
+                            }
+                            receiver = null;
+                        }
+                        //if we read data into a frame we just return immediately, even if there is more remaining
+                        //see https://issues.jboss.org/browse/UNDERTOW-410
+                        //basically if we don't do this we loose some message ordering semantics
+                        //as the second message may be processed before the first one
 
-                        if(isLastFrameReceived()) {
-                            handleLastFrame(newChannel);
+                        //this is problematic for HTTPS, where the read listener may also be invoked by a queued task
+                        //and not by the selector mechanism
+                        return null;
+                    }
+                    FrameHeaderData data = parseFrame(pooled.getBuffer());
+                    if (data != null) {
+                        PooledByteBuffer frameData;
+                        if (data.getFrameLength() >= pooled.getBuffer().remaining()) {
+                            frameDataRemaining = data.getFrameLength() - pooled.getBuffer().remaining();
+                            frameData = pooled.createView();
+                            pooled.getBuffer().position(pooled.getBuffer().limit());
+                        } else {
+                            frameData = pooled.createView((int) data.getFrameLength());
+                        }
+                        AbstractFramedStreamSourceChannel<?, ?, ?> existing = data.getExistingChannel();
+                        if (existing != null) {
+                            if (data.getFrameLength() > frameData.getBuffer().remaining()) {
+                                receiver = (R) existing;
+                            }
+                            existing.dataReady(data, frameData);
+                            if (isLastFrameReceived()) {
+                                handleLastFrame(existing);
+                            }
+                            return null;
+                        } else {
+                            boolean moreData = data.getFrameLength() > frameData.getBuffer().remaining();
+                            R newChannel = createChannel(data, frameData);
+                            if (newChannel != null) {
+                                if (moreData) {
+                                    receiver = newChannel;
+                                }
+
+                                if(isLastFrameReceived()) {
+                                    handleLastFrame(newChannel);
+                                }
+                            } else {
+                                frameData.close();
+                            }
+                            return newChannel;
                         }
                     } else {
-                        frameData.close();
+                        //we set partial read to true so the read listener knows not to immediately call receive again
+                        partialRead = true;
+                    }
+                    return null;
+                } catch (IOException|RuntimeException|Error e) {
+                    //something has code wrong with parsing, close the read side
+                    //we don't close the write side, as the underlying implementation will most likely want to send an error
+                    markReadsBroken(e);
+                    forceFree = true;
+                    throw e;
+                }finally {
+                    //if the receive caused the channel to break the close listener may be have been called
+                    //which will make readData null
+                    if (readData != null) {
+                        if (!pooled.getBuffer().hasRemaining() || forceFree) {
+                            if(pooled.getBuffer().capacity() < 1024 || forceFree) {
+                                //if there is less than 1k left we don't allow it to be re-aquired
+                                readData = null;
+                            }
+                            //even though this is freed we may un-free it if we get a new packet
+                            //this prevents many small reads resulting in a large number of allocated buffers
+                            pooled.close();
+
+                        }
+                    }
+                    if(requiresReinvoke) {
+                        if(readData != null && !readData.isFreed()) {
+                            if(readData.getBuffer().remaining() == reinvokeDataRemaining) {
+                                readData.close();
+                                readData = null;
+                                UndertowLogger.REQUEST_IO_LOGGER.debugf("Partial message read before connection close %s", this);
+                            }
+                        }
+                        channel.getSourceChannel().wakeupReads();
                     }
-                    return newChannel;
                 }
             }
-            return null;
-        } catch (IOException|RuntimeException|Error e) {
-            //something has code wrong with parsing, close the read side
-            //we don't close the write side, as the underlying implementation will most likely want to send an error
-            markReadsBroken(e);
-            forceFree = true;
-            throw e;
-        }finally {
-            //if the receive caused the channel to break the close listener may be have been called
-            //which will make readData null
-            if (readData != null) {
-                if (!pooled.getBuffer().hasRemaining() || forceFree) {
-                    if(pooled.getBuffer().limit() * 2 > pooled.getBuffer().capacity() || forceFree) {
-                        //if we have used more than half the buffer we don't allow it to be re-aquired
-                        readData = null;
-                    }
-                    //even though this is freed we may un-free it if we get a new packet
-                    //this prevents many small reads resulting in a large number of allocated buffers
-                    pooled.close();
-
-                }
+        } finally {
+            // read receivedMinusOne, and not readChannelDone
+            // to prevent lastDataRead being invoked twice in case of
+            // two concurrent receive invocations
+            if (receivedMinusOne) {
+                lastDataRead();
             }
         }
     }
@@ -512,10 +549,10 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
                         //we need to re-read in a sync block, to prevent races
                         expect = outstandingBuffersUpdater.get(this);
                         if (expect == maxQueuedBuffers) {
-                            if(UndertowLogger.REQUEST_IO_LOGGER.isTraceEnabled()) {
+                            if (UndertowLogger.REQUEST_IO_LOGGER.isTraceEnabled()) {
                                 UndertowLogger.REQUEST_IO_LOGGER.tracef("Suspending reads on %s due to too many outstanding buffers", this);
                             }
-                            channel.getSourceChannel().suspendReads();
+                            getIoThread().execute(new UpdateResumeState(null, true, null));
                             return null;
                         }
                     }
@@ -572,8 +609,8 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
         flushingSenders = true;
         try {
             int toSend = 0;
-            while (!newFrames.isEmpty()) {
-                S frame = newFrames.poll();
+            S frame;
+            while ((frame = newFrames.poll()) != null) {
                 frame.preWrite();
                 if (framePriority.insertFrame(frame, pendingFrames)) {
                     if (!heldFrames.isEmpty()) {
@@ -659,6 +696,10 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
                         channel.getSinkChannel().setWriteListener(ChannelListeners.flushingChannelListener(null, null));
                         channel.getSinkChannel().resumeWrites();
                     }
+                } else if (pendingFrames.size() > queuedFrameHighWaterMark) {
+                    new UpdateResumeState(null, null, true).run();
+                } else if (receivesSuspendedTooManyQueuedMessages && pendingFrames.size() < queuedFrameLowWaterMark) {
+                    new UpdateResumeState(null, null, false).run();
                 }
 
             } catch (IOException|RuntimeException|Error e) {
@@ -756,35 +797,16 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
      * Suspend the receive of new frames via {@link #receive()}
      */
     public synchronized void suspendReceives() {
-        receivesSuspended = true;
-        if (receiver == null) {
-            if(Thread.currentThread() == channel.getIoThread()) {
-                channel.getSourceChannel().suspendReads();
-            } else {
-                channel.getIoThread().execute(new Runnable() {
-                    @Override
-                    public void run() {
-                        channel.getSourceChannel().suspendReads();
-                    }
-                });
-            }
-        }
+        receivesSuspendedByUser = true;
+        getIoThread().execute(new UpdateResumeState(true, null, null));
     }
 
     /**
      * Resume the receive of new frames via {@link #receive()}
      */
     public synchronized void resumeReceives() {
-        receivesSuspended = false;if(Thread.currentThread() == channel.getIoThread()) {
-            doResume();
-        } else {
-            channel.getIoThread().execute(new Runnable() {
-                @Override
-                public void run() {
-                    doResume();
-                }
-            });
-        }
+        receivesSuspendedByUser = false;
+        getIoThread().execute(new UpdateResumeState(false, null, null));
     }
 
     private void doResume() {
@@ -796,7 +818,7 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
     }
 
     public boolean isReceivesResumed() {
-        return !receivesSuspended;
+        return !receivesSuspendedByUser;
     }
 
     /**
@@ -804,13 +826,15 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
      */
     @Override
     public void close() throws IOException {
-        if(UndertowLogger.REQUEST_IO_LOGGER.isTraceEnabled()) {
-            UndertowLogger.REQUEST_IO_LOGGER.tracef(new ClosedChannelException(), "Channel %s is being closed", this);
-        }
-        safeClose(channel);
-        if(readData != null) {
-            readData.close();
-            readData = null;
+        synchronized (this) {
+            if (UndertowLogger.REQUEST_IO_LOGGER.isTraceEnabled()) {
+                UndertowLogger.REQUEST_IO_LOGGER.tracef(new ClosedChannelException(), "Channel %s is being closed", this);
+            }
+            safeClose(channel);
+            if (readData != null) {
+                readData.close();
+                readData = null;
+            }
         }
         closeSubChannels();
     }
@@ -853,7 +877,6 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
     protected abstract void closeSubChannels();
 
 
-
     /**
      * Called when a sub channel fails to fulfil its contract, and leaves the channel in an inconsistent state.
      * <p>
@@ -925,23 +948,28 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
         @Override
         public void handleEvent(final StreamSourceChannel channel) {
             //clear the task queue before reading
-            while (!taskRunQueue.isEmpty()) {
-                taskRunQueue.poll().run();
+            Runnable runnable;
+            while ((runnable = taskRunQueue.poll()) != null) {
+                runnable.run();
             }
 
             final R receiver = AbstractFramedChannel.this.receiver;
-            if ((readChannelDone || receivesSuspended) && receiver == null) {
+            if ((readChannelDone || isReadsSuspended()) && receiver == null) {
                 channel.suspendReads();
                 return;
             } else {
                 ChannelListener listener = receiveSetter.get();
-                if(listener == null) {
+                if (listener == null) {
                     listener = DRAIN_LISTENER;
                 }
                 UndertowLogger.REQUEST_IO_LOGGER.tracef("Invoking receive listener", receiver);
                 ChannelListeners.invokeChannelListener(AbstractFramedChannel.this, listener);
             }
-            if (readData != null  && !readData.isFreed() && channel.isOpen()) {
+            final boolean partialRead;
+            synchronized (AbstractFramedChannel.this) {
+                partialRead = AbstractFramedChannel.this.partialRead;
+            }
+            if (readData != null && !readData.isFreed() && channel.isOpen() && !partialRead) {
                 try {
                     runInIoThread(new Runnable() {
                         @Override
@@ -953,9 +981,16 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
                     IoUtils.safeClose(AbstractFramedChannel.this);
                 }
             }
+            synchronized (AbstractFramedChannel.this) {
+                AbstractFramedChannel.this.partialRead = false;
+            }
         }
     }
 
+    private boolean isReadsSuspended() {
+        return receivesSuspendedByUser || receivesSuspendedTooManyBuffers || receivesSuspendedTooManyQueuedMessages;
+    }
+
     private class FrameWriteListener implements ChannelListener<StreamSinkChannel> {
         @Override
         public void handleEvent(final StreamSinkChannel channel) {
@@ -973,7 +1008,6 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
 
         @Override
         public void handleEvent(final CloseableChannel c) {
-
             if (Thread.currentThread() != c.getIoThread() && !c.getWorker().isShutdown()) {
                 runInIoThread(new Runnable() {
                     @Override
@@ -1096,9 +1130,6 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
     }
 
 
-
-
-
     protected ChannelExceptionHandler<SuspendableWriteChannel> writeExceptionHandler() {
         return new ChannelExceptionHandler<SuspendableWriteChannel>() {
             @Override
@@ -1119,4 +1150,35 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
     protected OptionMap getSettings() {
         return settings;
     }
+
+    private class UpdateResumeState implements Runnable {
+
+        private final Boolean user;
+        private final Boolean buffers;
+        private final Boolean frames;
+
+        private UpdateResumeState(Boolean user, Boolean buffers, Boolean frames) {
+            this.user = user;
+            this.buffers = buffers;
+            this.frames = frames;
+        }
+
+        @Override
+        public void run() {
+            if (user != null) {
+                receivesSuspendedByUser = user;
+            }
+            if (buffers != null) {
+                receivesSuspendedTooManyBuffers = buffers;
+            }
+            if (frames != null) {
+                receivesSuspendedTooManyQueuedMessages = frames;
+            }
+            if (receivesSuspendedByUser || receivesSuspendedTooManyQueuedMessages || receivesSuspendedTooManyBuffers) {
+                channel.getSourceChannel().suspendReads();
+            } else {
+                doResume();
+            }
+        }
+    }
 }
