@@ -109,7 +109,7 @@ final class HttpResponseConduit extends AbstractStreamSinkConduit<StreamSinkCond
      * Handles writing out the header data. It can also take a byte buffer of user
      * data, to enable both user data and headers to be written out in a single operation,
      * which has a noticeable performance impact.
-     * <p/>
+     * <p>
      * It is up to the caller to note the current position of this buffer before and after they
      * call this method, and use this to figure out how many bytes (if any) have been written.
      *
@@ -146,7 +146,7 @@ final class HttpResponseConduit extends AbstractStreamSinkConduit<StreamSinkCond
                         }
                         data[0] = byteBuffer;
                         System.arraycopy(userData, pos, data, 1, length);
-                        res = next.write(data, 0, data.length);
+                        res = next.write(data, 0, length + 1);
                     }
                     if (res == 0) {
                         return STATE_BUF_FLUSH;
@@ -294,8 +294,9 @@ final class HttpResponseConduit extends AbstractStreamSinkConduit<StreamSinkCond
         int length = string.length();
         for (int charIndex = 0; charIndex < length; charIndex++) {
             char c = string.charAt(charIndex);
-            if(c != '\r' && c != '\n') {
-                buffer.put((byte) c);
+            byte b = (byte) c;
+            if(b != '\r' && b != '\n') {
+                buffer.put(b);
             } else {
                 buffer.put((byte) ' ');
             }
