@@ -101,7 +101,7 @@ public class InMemorySessionManager implements SessionManager, SessionManagerSta
         this.sessions = new ConcurrentHashMap<>();
         this.maxSize = maxSessions;
         ConcurrentDirectDeque<String> evictionQueue = null;
-        if (maxSessions > 0) {
+        if (maxSessions > 0 && expireOldestUnusedSessionOnMax) {
             evictionQueue = ConcurrentDirectDeque.newInstance();
         }
         this.evictionQueue = evictionQueue;
@@ -140,7 +140,7 @@ public class InMemorySessionManager implements SessionManager, SessionManagerSta
 
     @Override
     public Session createSession(final HttpServerExchange serverExchange, final SessionConfig config) {
-        if (evictionQueue != null) {
+        if (maxSize > 0) {
             if(expireOldestUnusedSessionOnMax) {
                 while (sessions.size() >= maxSize && !evictionQueue.isEmpty()) {
 
@@ -214,7 +214,7 @@ public class InMemorySessionManager implements SessionManager, SessionManagerSta
         }
         String sessionId = config.findSessionId(serverExchange);
         InMemorySessionManager.SessionImpl session = (SessionImpl) getSession(sessionId);
-        if(session != null) {
+        if(session != null && serverExchange != null) {
             session.requestStarted(serverExchange);
         }
         return session;
@@ -298,6 +298,7 @@ public class InMemorySessionManager implements SessionManager, SessionManagerSta
         return maxSize;
     }
 
+    @Override
     public long getHighestSessionCount() {
         return highestSessionCount.get();
     }
@@ -560,6 +561,10 @@ public class InMemorySessionManager implements SessionManager, SessionManagerSta
             if(exchange != null) {
                 exchange.removeAttachment(sessionManager.NEW_SESSION);
             }
+            Object evictionToken = this.evictionToken;
+            if(evictionToken != null) {
+                sessionManager.evictionQueue.removeToken(evictionToken);
+            }
         }
 
         void invalidate(final HttpServerExchange exchange, SessionListener.SessionDestroyedReason reason) {
