@@ -33,6 +33,8 @@ import java.security.KeyStoreException;
 import java.security.NoSuchAlgorithmException;
 import java.security.UnrecoverableKeyException;
 import java.security.cert.CertificateException;
+import java.util.function.Supplier;
+
 import javax.net.ssl.KeyManager;
 import javax.net.ssl.KeyManagerFactory;
 import javax.net.ssl.SSLContext;
@@ -40,8 +42,6 @@ import javax.net.ssl.SSLEngine;
 import javax.net.ssl.TrustManager;
 import javax.net.ssl.TrustManagerFactory;
 
-import io.undertow.protocols.alpn.ALPNProvider;
-import io.undertow.protocols.alpn.JettyAlpnProvider;
 import org.junit.Assume;
 import org.junit.Ignore;
 import org.junit.runner.Description;
@@ -59,15 +59,21 @@ import org.xnio.ChannelListeners;
 import org.xnio.IoUtils;
 import org.xnio.OptionMap;
 import org.xnio.Options;
+import org.xnio.Sequence;
 import org.xnio.StreamConnection;
 import org.xnio.Xnio;
 import org.xnio.XnioWorker;
 import org.xnio.channels.AcceptingChannel;
 import org.xnio.ssl.XnioSsl;
+
 import io.undertow.UndertowLogger;
 import io.undertow.UndertowOptions;
 import io.undertow.connector.ByteBufferPool;
 import io.undertow.protocols.alpn.ALPNManager;
+import io.undertow.protocols.alpn.ALPNProvider;
+import io.undertow.protocols.alpn.JettyAlpnProvider;
+import io.undertow.protocols.ssl.SNIContextMatcher;
+import io.undertow.protocols.ssl.SNISSLContext;
 import io.undertow.protocols.ssl.UndertowXnioSsl;
 import io.undertow.security.impl.GSSAPIAuthenticationMechanism;
 import io.undertow.server.DefaultByteBufferPool;
@@ -155,7 +161,7 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
 
     private static KeyStore loadKeyStore(final String name) throws IOException {
         final InputStream stream = DefaultServer.class.getClassLoader().getResourceAsStream(name);
-        if(stream == null) {
+        if (stream == null) {
             throw new RuntimeException("Could not load keystore");
         }
         try {
@@ -171,6 +177,10 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
     }
 
     private static SSLContext createSSLContext(final KeyStore keyStore, final KeyStore trustStore, boolean client) throws IOException {
+        return createSSLContext(keyStore, trustStore, "TLSv1.2", client);
+    }
+
+    private static SSLContext createSSLContext(final KeyStore keyStore, final KeyStore trustStore, String protocol, boolean client) throws IOException {
         KeyManager[] keyManagers;
         try {
             KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
@@ -191,17 +201,23 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
 
         SSLContext sslContext;
         try {
-            if(openssl && !client) {
+            if (openssl && !client) {
                 sslContext = SSLContext.getInstance("openssl.TLS");
             } else {
-                sslContext = SSLContext.getInstance("TLS");
+                sslContext = SSLContext.getInstance(protocol);
             }
             sslContext.init(keyManagers, trustManagers, null);
         } catch (NoSuchAlgorithmException | KeyManagementException e) {
             throw new IOException("Unable to create and initialise the SSLContext", e);
         }
-
-        return sslContext;
+        if (!client) {
+            SNIContextMatcher matcher = new SNIContextMatcher.Builder().setDefaultContext(sslContext)
+                    .addMatch("localhost", sslContext)
+                    .build();
+            return new SNISSLContext(matcher);
+        } else {
+            return sslContext;
+        }
     }
 
     /**
@@ -236,6 +252,15 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
         return pool;
     }
 
+    public static Supplier<XnioWorker> getWorkerSupplier() {
+        return new Supplier<XnioWorker>() {
+            @Override
+            public XnioWorker get() {
+                return getWorker();
+            }
+        };
+    }
+
     @Override
     public Description getDescription() {
         return super.getDescription();
@@ -277,7 +302,7 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
     }
 
     private static void runInternal(final RunNotifier notifier) {
-        if(openssl && OPENSSL_FAILURE != null) {
+        if (openssl && OPENSSL_FAILURE != null) {
             throw new RuntimeException(OPENSSL_FAILURE);
         }
         if (first) {
@@ -460,7 +485,7 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
                 return;
             }
         }
-        if(h2 || h2c || ajp || h2cUpgrade) {
+        if (h2 || h2c || ajp || h2cUpgrade) {
             //h2c-upgrade we still allow HTTP1
             HttpOneOnly httpOneOnly = method.getAnnotation(HttpOneOnly.class);
             if (httpOneOnly == null) {
@@ -544,7 +569,7 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
     public static void setRootHandler(HttpHandler handler) {
         if ((isProxy()) && !ajp) {
             //if we are testing HTTP proxy we always add the SSLHeaderHandler
-            //this allows the SSL information to be propagated to be backend
+            //this allows the SSL information to be propagated to the backend
             handler = new SSLHeaderHandler(new ProxyPeerAddressHandler(handler));
         }
         if (dump) {
@@ -556,7 +581,7 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
 
     /**
      * When using the default SSL settings returns the corresponding client context.
-     * <p/>
+     * <p>
      * If a test case is initialising a custom server side SSLContext then the test case will be responsible for creating it's
      * own client side.
      *
@@ -571,7 +596,7 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
 
     /**
      * Start the SSL server using the default settings.
-     * <p/>
+     * <p>
      * The default settings initialise a server with a key for 'localhost' and a trust store containing the certificate of a
      * single client, the client authentication mode is set to 'REQUESTED' to optionally allow progression to CLIENT-CERT
      * authentication.
@@ -580,12 +605,16 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
         SSLContext serverContext = getServerSslContext();
         getClientSSLContext();
 
-        startSSLServer(serverContext, OptionMap.create(SSL_CLIENT_AUTH_MODE, REQUESTED));
+        startSSLServer(serverContext, OptionMap.create(SSL_CLIENT_AUTH_MODE, REQUESTED, Options.SSL_ENABLED_PROTOCOLS, Sequence.of("TLSv1.2")));
     }
 
     public static SSLContext createClientSslContext() {
+        return createClientSslContext("TLSv1.2");
+    }
+
+    public static SSLContext createClientSslContext(String protocol) {
         try {
-            return createSSLContext(loadKeyStore(CLIENT_KEY_STORE), loadKeyStore(CLIENT_TRUST_STORE), true);
+            return createSSLContext(loadKeyStore(CLIENT_KEY_STORE), loadKeyStore(CLIENT_TRUST_STORE), protocol, true);
         } catch (IOException e) {
             throw new RuntimeException(e);
         }
@@ -601,19 +630,18 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
 
     /**
      * Start the SSL server using the default ssl context and the provided option map
-     * <p/>
+     * <p>
      * The default settings initialise a server with a key for 'localhost' and a trust store containing the certificate of a
      * single client. Client cert mode is not set by default
      */
     public static void startSSLServer(OptionMap optionMap) throws IOException {
-        SSLContext serverContext = getServerSslContext();
         clientSslContext = createClientSslContext();
         startSSLServer(optionMap, proxyAcceptListener != null ? proxyAcceptListener : acceptListener);
     }
 
     /**
      * Start the SSL server using the default ssl context and the provided option map
-     * <p/>
+     * <p>
      * The default settings initialise a server with a key for 'localhost' and a trust store containing the certificate of a
      * single client. Client cert mode is not set by default
      */
@@ -681,6 +709,11 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
             sslServer = null;
         }
         clientSslContext = null;
+        if(proxyOpenListener != null) {
+            proxyOpenListener.closeConnections();
+        } else {
+            openListener.closeConnections();
+        }
     }
 
     public static String getHostAddress(String serverName) {
@@ -719,7 +752,11 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
             builder.set(UndertowOptions.ENABLE_HTTP2, true);
         }
         openListener.setUndertowOptions(builder.getMap());
-        if(loadBalancingProxyClient != null) {
+        openListener.closeConnections();
+        if(proxyOpenListener != null) {
+            proxyOpenListener.closeConnections();
+        }
+        if (loadBalancingProxyClient != null) {
             loadBalancingProxyClient.closeCurrentConnections();
         }
     }
@@ -764,7 +801,7 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
     /**
      * The root handler is tied to the connection, and AJP can re-use connections for different tests, so we
      * use a delegating handler to chance the next handler after the root.
-     * <p/>
+     * <p>
      * TODO: should we re-read the root handler for every request?
      */
     private static final class DelegatingHandler implements HttpHandler {
@@ -781,9 +818,11 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
 
     private static boolean isAlpnEnabled() {
         if (alpnEnabled == null) {
-            SSLEngine engine = getServerSslContext().createSSLEngine();
+            //we use the client context, as the server one is wrapped by a SNISSLEngine
+            //so we can't tell that ALPN is enabled or now
+            SSLEngine engine = getClientSSLContext().createSSLEngine();
             ALPNProvider provider = ALPNManager.INSTANCE.getProvider(engine);
-            if(provider instanceof JettyAlpnProvider) {
+            if (provider instanceof JettyAlpnProvider) {
                 alpnEnabled = System.getProperty("alpn-boot-string") != null;
             } else {
                 alpnEnabled = provider != null;
