@@ -71,6 +71,7 @@ import io.undertow.servlet.api.SessionPersistenceManager;
 import io.undertow.servlet.api.ThreadSetupHandler;
 import io.undertow.servlet.api.WebResourceCollection;
 import io.undertow.servlet.handlers.CrawlerSessionManagerHandler;
+import io.undertow.servlet.handlers.RedirectDirHandler;
 import io.undertow.servlet.handlers.ServletDispatchingHandler;
 import io.undertow.servlet.handlers.ServletHandler;
 import io.undertow.servlet.handlers.ServletInitialHandler;
@@ -90,8 +91,9 @@ import io.undertow.util.MimeMappings;
 
 import javax.servlet.ServletContainerInitializer;
 import javax.servlet.ServletContext;
+import javax.servlet.ServletContextEvent;
+import javax.servlet.ServletContextListener;
 import javax.servlet.ServletException;
-
 import java.io.File;
 import java.nio.charset.Charset;
 import java.util.ArrayList;
@@ -163,6 +165,16 @@ public class DeploymentManagerImpl implements DeploymentManager {
         if (deploymentInfo.getDefaultEncoding() != null) {
             deployment.setDefaultCharset(Charset.forName(deploymentInfo.getDefaultEncoding()));
         }
+        if(deploymentInfo.getDefaultRequestEncoding() != null) {
+            deployment.setDefaultRequestCharset(Charset.forName(deploymentInfo.getDefaultRequestEncoding()));
+        } else if (deploymentInfo.getDefaultEncoding() != null) {
+            deployment.setDefaultRequestCharset(Charset.forName(deploymentInfo.getDefaultEncoding()));
+        }
+        if(deploymentInfo.getDefaultResponseEncoding() != null) {
+            deployment.setDefaultResponseCharset(Charset.forName(deploymentInfo.getDefaultResponseEncoding()));
+        } else if (deploymentInfo.getDefaultEncoding() != null) {
+            deployment.setDefaultResponseCharset(Charset.forName(deploymentInfo.getDefaultEncoding()));
+        }
 
         handleDeploymentSessionConfig(deploymentInfo, servletContext);
 
@@ -207,6 +219,7 @@ public class DeploymentManagerImpl implements DeploymentManager {
 
                     HttpHandler wrappedHandlers = ServletDispatchingHandler.INSTANCE;
                     wrappedHandlers = wrapHandlers(wrappedHandlers, deploymentInfo.getInnerHandlerChainWrappers());
+                    wrappedHandlers = new RedirectDirHandler(wrappedHandlers, deployment.getServletPaths());
                     if(!deploymentInfo.isSecurityDisabled()) {
                         HttpHandler securityHandler = setupSecurityHandlers(wrappedHandlers);
                         wrappedHandlers = new PredicateHandler(DispatcherTypePredicate.REQUEST, securityHandler, wrappedHandlers);
@@ -243,6 +256,9 @@ public class DeploymentManagerImpl implements DeploymentManager {
         //any problems with the paths won't get detected until the data is initialize
         //so we force initialization here
         deployment.getServletPaths().initData();
+        for(ServletContextListener listener : deploymentInfo.getDeploymentCompleteListeners()) {
+            listener.contextInitialized(new ServletContextEvent(servletContext));
+        }
         state = State.DEPLOYED;
     }
 
@@ -256,27 +272,27 @@ public class DeploymentManagerImpl implements DeploymentManager {
     }
 
     private void handleExtensions(final DeploymentInfo deploymentInfo, final ServletContextImpl servletContext) {
-        Set<Class<?>> loadedExtensions = new HashSet<>();
+        Set<String> loadedExtensions = new HashSet<>();
 
         for (ServletExtension extension : ServiceLoader.load(ServletExtension.class, deploymentInfo.getClassLoader())) {
-            loadedExtensions.add(extension.getClass());
+            loadedExtensions.add(extension.getClass().getName());
             extension.handleDeployment(deploymentInfo, servletContext);
         }
 
-        if (!ServletExtension.class.getClassLoader().equals(deploymentInfo.getClassLoader())) {
+        if (ServletExtension.class.getClassLoader() != null && !ServletExtension.class.getClassLoader().equals(deploymentInfo.getClassLoader())) {
             for (ServletExtension extension : ServiceLoader.load(ServletExtension.class)) {
 
                 // Note: If the CLs are different, but can the see the same extensions and extension might get loaded
                 // and thus instantiated twice, but the handleDeployment() is executed only once.
 
-                if (!loadedExtensions.contains(extension.getClass())) {
+                if (!loadedExtensions.contains(extension.getClass().getName())) {
                     extension.handleDeployment(deploymentInfo, servletContext);
                 }
             }
         }
 
         for (ServletExtension extension : ServletExtensionHolder.getServletExtensions()) {
-            if (!loadedExtensions.contains(extension.getClass())) {
+            if (!loadedExtensions.contains(extension.getClass().getName())) {
                 extension.handleDeployment(deploymentInfo, servletContext);
             }
         }
@@ -288,7 +304,7 @@ public class DeploymentManagerImpl implements DeploymentManager {
 
     /**
      * sets up the outer security handlers.
-     * <p/>
+     * <p>
      * the handler that actually performs the access check happens later in the chain, it is not setup here
      *
      * @param initialHandler The handler to wrap with security handlers
@@ -326,22 +342,22 @@ public class DeploymentManagerImpl implements DeploymentManager {
             final Map<String, AuthenticationMechanismFactory> factoryMap = new HashMap<>(deploymentInfo.getAuthenticationMechanisms());
             final IdentityManager identityManager = deploymentInfo.getIdentityManager();
             if(!factoryMap.containsKey(BASIC_AUTH)) {
-                factoryMap.put(BASIC_AUTH, new BasicAuthenticationMechanism.Factory(identityManager));
+                factoryMap.put(BASIC_AUTH, BasicAuthenticationMechanism.FACTORY);
             }
             if(!factoryMap.containsKey(FORM_AUTH)) {
-                factoryMap.put(FORM_AUTH, new ServletFormAuthenticationMechanism.Factory(identityManager));
+                factoryMap.put(FORM_AUTH, ServletFormAuthenticationMechanism.FACTORY);
             }
             if(!factoryMap.containsKey(DIGEST_AUTH)) {
-                factoryMap.put(DIGEST_AUTH, new DigestAuthenticationMechanism.Factory(identityManager));
+                factoryMap.put(DIGEST_AUTH, DigestAuthenticationMechanism.FACTORY);
             }
             if(!factoryMap.containsKey(CLIENT_CERT_AUTH)) {
-                factoryMap.put(CLIENT_CERT_AUTH, new ClientCertAuthenticationMechanism.Factory(identityManager));
+                factoryMap.put(CLIENT_CERT_AUTH, ClientCertAuthenticationMechanism.FACTORY);
             }
             if(!factoryMap.containsKey(ExternalAuthenticationMechanism.NAME)) {
-                factoryMap.put(ExternalAuthenticationMechanism.NAME, new ExternalAuthenticationMechanism.Factory(identityManager));
+                factoryMap.put(ExternalAuthenticationMechanism.NAME, ExternalAuthenticationMechanism.FACTORY);
             }
             if(!factoryMap.containsKey(GenericHeaderAuthenticationMechanism.NAME)) {
-                factoryMap.put(GenericHeaderAuthenticationMechanism.NAME, new GenericHeaderAuthenticationMechanism.Factory(identityManager));
+                factoryMap.put(GenericHeaderAuthenticationMechanism.NAME, GenericHeaderAuthenticationMechanism.FACTORY);
             }
             List<AuthenticationMechanism> authenticationMechanisms = new LinkedList<>();
 
@@ -352,8 +368,12 @@ public class DeploymentManagerImpl implements DeploymentManager {
 
                 //we don't allow multipart requests, and use the default encoding when it's set
                 FormEncodedDataDefinition formEncodedDataDefinition = new FormEncodedDataDefinition();
-                if (deploymentInfo.getDefaultEncoding() != null) {
-                    formEncodedDataDefinition.setDefaultEncoding(deploymentInfo.getDefaultEncoding());
+                String reqEncoding = deploymentInfo.getDefaultRequestEncoding();
+                if(reqEncoding == null) {
+                    reqEncoding = deploymentInfo.getDefaultEncoding();
+                }
+                if (reqEncoding != null) {
+                    formEncodedDataDefinition.setDefaultEncoding(reqEncoding);
                 }
                 FormParserFactory parser = FormParserFactory.builder(false)
                         .addParser(formEncodedDataDefinition)
@@ -388,7 +408,7 @@ public class DeploymentManagerImpl implements DeploymentManager {
                     name = name.equals(DIGEST_AUTH) ? DIGEST_AUTH : name;
                     name = name.equals(CLIENT_CERT_AUTH) ? CLIENT_CERT_AUTH : name;
 
-                    authenticationMechanisms.add(factory.create(name, parser, properties));
+                    authenticationMechanisms.add(factory.create(name, identityManager, parser, properties));
                 }
             }
 
@@ -516,7 +536,7 @@ public class DeploymentManagerImpl implements DeploymentManager {
     private ApplicationListeners createListeners() {
         final List<ManagedListener> managedListeners = new ArrayList<>();
         for (final ListenerInfo listener : deployment.getDeploymentInfo().getListeners()) {
-            managedListeners.add(new ManagedListener(listener, false));
+            managedListeners.add(new ManagedListener(listener, listener.isProgramatic()));
         }
         return new ApplicationListeners(managedListeners, deployment.getServletContext());
     }
@@ -592,8 +612,8 @@ public class DeploymentManagerImpl implements DeploymentManager {
                     for (Lifecycle object : deployment.getLifecycleObjects()) {
                         try {
                             object.stop();
-                        } catch (Exception e) {
-                            UndertowServletLogger.ROOT_LOGGER.failedToDestroy(object, e);
+                        } catch (Throwable t) {
+                            UndertowServletLogger.ROOT_LOGGER.failedToDestroy(object, t);
                         }
                     }
                     deployment.getSessionManager().stop();
@@ -646,6 +666,13 @@ public class DeploymentManagerImpl implements DeploymentManager {
             deployment.createThreadSetupAction(new ThreadSetupHandler.Action<Void, Object>() {
                 @Override
                 public Void call(HttpServerExchange exchange, Object ignore) throws ServletException {
+                    for(ServletContextListener listener : deployment.getDeploymentInfo().getDeploymentCompleteListeners()) {
+                        try {
+                            listener.contextDestroyed(new ServletContextEvent(deployment.getServletContext()));
+                        } catch (Throwable t) {
+                            UndertowServletLogger.REQUEST_LOGGER.failedToDestroy(listener, t);
+                        }
+                    }
                     deployment.destroy();
                     deployment = null;
                     state = State.UNDEPLOYED;
