@@ -31,6 +31,7 @@ import java.util.List;
 import java.util.Map;
 import java.util.Set;
 import javax.servlet.DispatcherType;
+import javax.servlet.http.MappingMatch;
 
 import io.undertow.server.HandlerWrapper;
 import io.undertow.server.HttpHandler;
@@ -48,7 +49,6 @@ import io.undertow.servlet.core.ManagedServlet;
 import io.undertow.servlet.core.ManagedServlets;
 import io.undertow.servlet.handlers.security.ServletSecurityRoleHandler;
 
-
 /**
  * Facade around {@link ServletPathMatchesData}. This facade is responsible for re-generating the matches if anything changes.
  *
@@ -190,10 +190,10 @@ public class ServletPathMatches {
     /**
      * Sets up the handlers in the servlet chain. We setup a chain for every path + extension match possibility.
      * (i.e. if there a m path mappings and n extension mappings we have n*m chains).
-     * <p/>
+     * <p>
      * If a chain consists of only the default servlet then we add it as an async handler, so that resources can be
      * served up directly without using blocking operations.
-     * <p/>
+     * <p>
      * TODO: this logic is a bit convoluted at the moment, we should look at simplifying it
      */
     private ServletPathMatchesData setupServletChains() {
@@ -324,7 +324,7 @@ public class ServletPathMatches {
             if (path.endsWith("/*")) {
                 String prefix = path.substring(0, path.length() - 2);
                 //add the default non-extension match
-                builder.addPrefixMatch(prefix, createHandler(deploymentInfo, targetServletMatch.handler, noExtension, targetServletMatch.matchedPath, targetServletMatch.defaultServlet), targetServletMatch.defaultServlet || targetServletMatch.handler.getManagedServlet().getServletInfo().isRequireWelcomeFileMapping());
+                builder.addPrefixMatch(prefix, createHandler(deploymentInfo, targetServletMatch.handler, noExtension, targetServletMatch.matchedPath, targetServletMatch.defaultServlet, targetServletMatch.mappingMatch, targetServletMatch.userPath), targetServletMatch.defaultServlet || targetServletMatch.handler.getManagedServlet().getServletInfo().isRequireWelcomeFileMapping());
 
                 //build up the chain for each non-extension match
                 for (Map.Entry<String, Map<DispatcherType, List<ManagedFilter>>> entry : extension.entrySet()) {
@@ -340,11 +340,11 @@ public class ServletPathMatches {
                     if (!entry.getValue().isEmpty()) {
                         handler = new FilterHandler(entry.getValue(), deploymentInfo.isAllowNonStandardWrappers(), handler);
                     }
-                    builder.addExtensionMatch(prefix, entry.getKey(), servletChain(handler, pathServlet.getManagedServlet(), entry.getValue(), pathMatch, deploymentInfo, defaultServletMatch));
+                    builder.addExtensionMatch(prefix, entry.getKey(), servletChain(handler, pathServlet.getManagedServlet(), entry.getValue(), pathMatch, deploymentInfo, defaultServletMatch, defaultServletMatch ? MappingMatch.DEFAULT : MappingMatch.EXTENSION, defaultServletMatch ? "/" : "*." + entry.getKey()));
                 }
             } else if (path.isEmpty()) {
                 //the context root match
-                builder.addExactMatch("/", createHandler(deploymentInfo, targetServletMatch.handler, noExtension, targetServletMatch.matchedPath, targetServletMatch.defaultServlet));
+                builder.addExactMatch("/", createHandler(deploymentInfo, targetServletMatch.handler, noExtension, targetServletMatch.matchedPath, targetServletMatch.defaultServlet, targetServletMatch.mappingMatch, targetServletMatch.userPath));
             } else {
                 //we need to check for an extension match, so paths like /exact.txt will have the correct filter applied
                 int lastSegmentIndex = path.lastIndexOf('/');
@@ -358,12 +358,12 @@ public class ServletPathMatches {
                     String ext = lastSegment.substring(lastSegment.lastIndexOf('.') + 1);
                     if (extension.containsKey(ext)) {
                         Map<DispatcherType, List<ManagedFilter>> extMap = extension.get(ext);
-                        builder.addExactMatch(path, createHandler(deploymentInfo, targetServletMatch.handler, extMap, targetServletMatch.matchedPath, targetServletMatch.defaultServlet));
+                        builder.addExactMatch(path, createHandler(deploymentInfo, targetServletMatch.handler, extMap, targetServletMatch.matchedPath, targetServletMatch.defaultServlet, targetServletMatch.mappingMatch, targetServletMatch.userPath));
                     } else {
-                        builder.addExactMatch(path, createHandler(deploymentInfo, targetServletMatch.handler, noExtension, targetServletMatch.matchedPath, targetServletMatch.defaultServlet));
+                        builder.addExactMatch(path, createHandler(deploymentInfo, targetServletMatch.handler, noExtension, targetServletMatch.matchedPath, targetServletMatch.defaultServlet, targetServletMatch.mappingMatch, targetServletMatch.userPath));
                     }
                 } else {
-                    builder.addExactMatch(path, createHandler(deploymentInfo, targetServletMatch.handler, noExtension, targetServletMatch.matchedPath, targetServletMatch.defaultServlet));
+                    builder.addExactMatch(path, createHandler(deploymentInfo, targetServletMatch.handler, noExtension, targetServletMatch.matchedPath, targetServletMatch.defaultServlet, targetServletMatch.mappingMatch, targetServletMatch.userPath));
                 }
 
             }
@@ -382,22 +382,22 @@ public class ServletPathMatches {
                 }
             }
             if (filtersByDispatcher.isEmpty()) {
-                builder.addNameMatch(entry.getKey(), servletChain(entry.getValue(), entry.getValue().getManagedServlet(), filtersByDispatcher, null, deploymentInfo, false));
+                builder.addNameMatch(entry.getKey(), servletChain(entry.getValue(), entry.getValue().getManagedServlet(), filtersByDispatcher, null, deploymentInfo, false, MappingMatch.EXACT, ""));
             } else {
-                builder.addNameMatch(entry.getKey(), servletChain(new FilterHandler(filtersByDispatcher, deploymentInfo.isAllowNonStandardWrappers(), entry.getValue()), entry.getValue().getManagedServlet(), filtersByDispatcher, null, deploymentInfo, false));
+                builder.addNameMatch(entry.getKey(), servletChain(new FilterHandler(filtersByDispatcher, deploymentInfo.isAllowNonStandardWrappers(), entry.getValue()), entry.getValue().getManagedServlet(), filtersByDispatcher, null, deploymentInfo, false, MappingMatch.EXACT, ""));
             }
         }
 
         return builder.build();
     }
 
-    private ServletChain createHandler(final DeploymentInfo deploymentInfo, final ServletHandler targetServlet, final Map<DispatcherType, List<ManagedFilter>> noExtension, final String servletPath, final boolean defaultServlet) {
+    private ServletChain createHandler(final DeploymentInfo deploymentInfo, final ServletHandler targetServlet, final Map<DispatcherType, List<ManagedFilter>> noExtension, final String servletPath, final boolean defaultServlet, MappingMatch mappingMatch, String pattern) {
         final ServletChain initialHandler;
         if (noExtension.isEmpty()) {
-            initialHandler = servletChain(targetServlet, targetServlet.getManagedServlet(), noExtension, servletPath, deploymentInfo, defaultServlet);
+            initialHandler = servletChain(targetServlet, targetServlet.getManagedServlet(), noExtension, servletPath, deploymentInfo, defaultServlet, mappingMatch, pattern);
         } else {
             FilterHandler handler = new FilterHandler(noExtension, deploymentInfo.isAllowNonStandardWrappers(), targetServlet);
-            initialHandler = servletChain(handler, targetServlet.getManagedServlet(), noExtension, servletPath, deploymentInfo, defaultServlet);
+            initialHandler = servletChain(handler, targetServlet.getManagedServlet(), noExtension, servletPath, deploymentInfo, defaultServlet, mappingMatch, pattern);
         }
         return initialHandler;
     }
@@ -406,13 +406,17 @@ public class ServletPathMatches {
         if (pathServlets.containsKey(path)) {
             if (path.endsWith("/*")) {
                 final String base = path.substring(0, path.length() - 2);
-                return new MatchData(pathServlets.get(path), base, false);
+                return new MatchData(pathServlets.get(path), base, path, MappingMatch.PATH, false);
             } else {
-                return new MatchData(pathServlets.get(path), path, false);
+                if(path.equals("/")) {
+                    return new MatchData(pathServlets.get(path), path, "", MappingMatch.CONTEXT_ROOT, false);
+                }
+                return new MatchData(pathServlets.get(path), path, path, MappingMatch.EXACT, false);
             }
         }
         String match = null;
         ServletHandler servlet = null;
+        String userPath = "";
         for (final Map.Entry<String, ServletHandler> entry : pathServlets.entrySet()) {
             String key = entry.getKey();
             if (key.endsWith("/*")) {
@@ -421,23 +425,24 @@ public class ServletPathMatches {
                     if (path.startsWith(base) || path.equals(base.substring(0, base.length() - 1))) {
                         match = base.substring(0, base.length() - 1);
                         servlet = entry.getValue();
+                        userPath = key;
                     }
                 }
             }
         }
         if (servlet != null) {
-            return new MatchData(servlet, match, false);
+            return new MatchData(servlet, match, userPath, MappingMatch.PATH, false);
         }
         int index = path.lastIndexOf('.');
         if (index != -1) {
             String ext = path.substring(index + 1);
             servlet = extensionServlets.get(ext);
             if (servlet != null) {
-                return new MatchData(servlet, null, false);
+                return new MatchData(servlet, null, "*." + ext, MappingMatch.EXTENSION, false);
             }
         }
 
-        return new MatchData(defaultServlet, null, true);
+        return new MatchData(defaultServlet, null, "/", MappingMatch.DEFAULT, true);
     }
 
     private static boolean isFilterApplicable(final String path, final String filterPath) {
@@ -452,7 +457,8 @@ public class ServletPathMatches {
         }
         if (modifiedPath.endsWith("/*")) {
             String baseFilterPath = modifiedPath.substring(0, modifiedPath.length() - 1);
-            return path.startsWith(baseFilterPath);
+            String exactFilterPath = modifiedPath.substring(0, modifiedPath.length() - 2);
+            return path.startsWith(baseFilterPath) || path.equals(exactFilterPath);
         } else {
             return modifiedPath.equals(path);
         }
@@ -466,13 +472,13 @@ public class ServletPathMatches {
         list.add(value);
     }
 
-    private static ServletChain servletChain(HttpHandler next, final ManagedServlet managedServlet, Map<DispatcherType, List<ManagedFilter>> filters, final String servletPath, final DeploymentInfo deploymentInfo, boolean defaultServlet) {
+    private static ServletChain servletChain(HttpHandler next, final ManagedServlet managedServlet, Map<DispatcherType, List<ManagedFilter>> filters, final String servletPath, final DeploymentInfo deploymentInfo, boolean defaultServlet, MappingMatch mappingMatch, String pattern) {
         HttpHandler servletHandler = next;
         if(!deploymentInfo.isSecurityDisabled()) {
             servletHandler = new ServletSecurityRoleHandler(servletHandler, deploymentInfo.getAuthorizationManager());
         }
         servletHandler = wrapHandlers(servletHandler, managedServlet.getServletInfo().getHandlerChainWrappers());
-        return new ServletChain(servletHandler, managedServlet, servletPath, defaultServlet, filters);
+        return new ServletChain(servletHandler, managedServlet, servletPath, defaultServlet, mappingMatch, pattern, filters);
     }
 
     private static HttpHandler wrapHandlers(final HttpHandler wrapee, final List<HandlerWrapper> wrappers) {
@@ -486,11 +492,15 @@ public class ServletPathMatches {
     private static class MatchData {
         final ServletHandler handler;
         final String matchedPath;
+        final String userPath;
+        final MappingMatch mappingMatch;
         final boolean defaultServlet;
 
-        private MatchData(final ServletHandler handler, final String matchedPath, boolean defaultServlet) {
+        private MatchData(final ServletHandler handler, final String matchedPath, String userPath, MappingMatch mappingMatch, boolean defaultServlet) {
             this.handler = handler;
             this.matchedPath = matchedPath;
+            this.userPath = userPath;
+            this.mappingMatch = mappingMatch;
             this.defaultServlet = defaultServlet;
         }
     }
