@@ -25,6 +25,7 @@ import io.undertow.server.RequestTooBigException;
 import io.undertow.server.handlers.form.FormData;
 import io.undertow.server.handlers.form.FormDataParser;
 import io.undertow.server.handlers.form.MultiPartParserDefinition;
+import io.undertow.server.protocol.http.HttpAttachments;
 import io.undertow.server.session.Session;
 import io.undertow.server.session.SessionConfig;
 import io.undertow.servlet.UndertowServletMessages;
@@ -43,26 +44,11 @@ import io.undertow.util.AttachmentKey;
 import io.undertow.util.CanonicalPathUtils;
 import io.undertow.util.DateUtils;
 import io.undertow.util.HeaderMap;
+import io.undertow.util.HeaderValues;
 import io.undertow.util.Headers;
 import io.undertow.util.HttpString;
 import io.undertow.util.LocaleUtils;
-import io.undertow.util.Methods;
 
-import javax.servlet.AsyncContext;
-import javax.servlet.DispatcherType;
-import javax.servlet.RequestDispatcher;
-import javax.servlet.ServletException;
-import javax.servlet.ServletInputStream;
-import javax.servlet.ServletRequest;
-import javax.servlet.ServletRequestWrapper;
-import javax.servlet.ServletResponse;
-import javax.servlet.ServletResponseWrapper;
-import javax.servlet.http.Cookie;
-import javax.servlet.http.HttpServletRequest;
-import javax.servlet.http.HttpServletResponse;
-import javax.servlet.http.HttpSession;
-import javax.servlet.http.HttpUpgradeHandler;
-import javax.servlet.http.Part;
 import java.io.BufferedReader;
 import java.io.IOException;
 import java.io.InputStreamReader;
@@ -70,6 +56,7 @@ import java.io.UnsupportedEncodingException;
 import java.net.InetAddress;
 import java.net.InetSocketAddress;
 import java.nio.charset.Charset;
+import java.nio.charset.StandardCharsets;
 import java.nio.charset.UnsupportedCharsetException;
 import java.security.AccessController;
 import java.security.Principal;
@@ -86,8 +73,24 @@ import java.util.List;
 import java.util.Locale;
 import java.util.Map;
 import java.util.Set;
-
+import javax.servlet.AsyncContext;
+import javax.servlet.DispatcherType;
 import javax.servlet.MultipartConfigElement;
+import javax.servlet.RequestDispatcher;
+import javax.servlet.ServletException;
+import javax.servlet.ServletInputStream;
+import javax.servlet.ServletRequest;
+import javax.servlet.ServletRequestWrapper;
+import javax.servlet.ServletResponse;
+import javax.servlet.ServletResponseWrapper;
+import javax.servlet.http.Cookie;
+import javax.servlet.http.HttpServletMapping;
+import javax.servlet.http.HttpServletRequest;
+import javax.servlet.http.HttpServletResponse;
+import javax.servlet.http.HttpSession;
+import javax.servlet.http.HttpUpgradeHandler;
+import javax.servlet.http.Part;
+import javax.servlet.http.PushBuilder;
 
 /**
  * The http servlet request implementation. This class is not thread safe
@@ -99,6 +102,9 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
     @Deprecated
     public static final AttachmentKey<Boolean> SECURE_REQUEST = HttpServerExchange.SECURE_REQUEST;
 
+    static final AttachmentKey<Boolean> REQUESTED_SESSION_ID_SET = AttachmentKey.create(Boolean.class);
+    static final AttachmentKey<String> REQUESTED_SESSION_ID = AttachmentKey.create(String.class);
+
     private final HttpServerExchange exchange;
     private final ServletContextImpl originalServletContext;
     private ServletContextImpl servletContext;
@@ -111,7 +117,6 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
     private Cookie[] cookies;
     private List<Part> parts = null;
     private volatile boolean asyncStarted = false;
-    private volatile boolean asyncCancelled = false;
     private volatile AsyncContextImpl asyncContext = null;
     private Map<String, Deque<String>> queryParameters;
     private FormData parsedFormData;
@@ -221,6 +226,43 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         return new IteratorEnumeration<>(headers.iterator());
     }
 
+    @Override
+    public HttpServletMapping getHttpServletMapping() {
+        ServletRequestContext src = exchange.getAttachment(ServletRequestContext.ATTACHMENT_KEY);
+        ServletPathMatch match = src.getOriginalServletPathMatch();
+        if(getDispatcherType() == DispatcherType.FORWARD) {
+            match = src.getServletPathMatch();
+        }
+        String matchValue;
+        switch (match.getMappingMatch()) {
+            case EXACT:
+                matchValue = match.getMatched();
+                if(matchValue.startsWith("/")) {
+                    matchValue = matchValue.substring(1);
+                }
+                break;
+            case DEFAULT:
+            case CONTEXT_ROOT:
+                matchValue = "";
+                break;
+            case PATH:
+                matchValue = match.getRemaining();
+                if(matchValue.startsWith("/")) {
+                    matchValue = matchValue.substring(1);
+                }
+                break;
+            case EXTENSION:
+                matchValue = match.getMatched().substring(0, match.getMatched().length() - match.getMatchString().length() + 1);
+                if(matchValue.startsWith("/")) {
+                    matchValue = matchValue.substring(1);
+                }
+                break;
+            default:
+                matchValue = match.getRemaining();
+        }
+        return new MappingImpl(matchValue, match.getMatchString(), match.getMappingMatch(), match.getServletChain().getManagedServlet().getServletInfo().getName());
+    }
+
     @Override
     public int getIntHeader(final String name) {
         String header = getHeader(name);
@@ -308,6 +350,10 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
 
     @Override
     public String getRequestedSessionId() {
+        Boolean isRequestedSessionIdSaved = exchange.getAttachment(REQUESTED_SESSION_ID_SET);
+        if (isRequestedSessionIdSaved != null && isRequestedSessionIdSaved) {
+            return exchange.getAttachment(REQUESTED_SESSION_ID);
+        }
         SessionConfig config = originalServletContext.getSessionConfig();
         if(config instanceof ServletContextImpl.ServletContextSessionConfig) {
             return ((ServletContextImpl.ServletContextSessionConfig)config).getDelegate().findSessionId(exchange);
@@ -562,11 +608,14 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         if (characterEncodingFromHeader != null) {
             return characterEncodingFromHeader;
         }
-
+        // first check, web-app context level default request encoding
+        if (servletContext.getDeployment().getDeploymentInfo().getDefaultRequestEncoding() != null) {
+            return servletContext.getDeployment().getDeploymentInfo().getDefaultRequestEncoding();
+        }
+        // now check the container level default encoding
         if (servletContext.getDeployment().getDeploymentInfo().getDefaultEncoding() != null) {
-            return servletContext.getDeployment().getDefaultCharset().name();
+            return servletContext.getDeployment().getDeploymentInfo().getDefaultEncoding();
         }
-
         return null;
     }
 
@@ -665,7 +714,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
             final FormData parsedFormData = parseFormData();
             if (parsedFormData != null) {
                 FormData.FormValue res = parsedFormData.getFirst(name);
-                if (res == null || res.isFile()) {
+                if (res == null || res.isFileItem()) {
                     return null;
                 } else {
                     return res.getValue();
@@ -682,17 +731,15 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
             queryParameters = exchange.getQueryParameters();
         }
         final Set<String> parameterNames = new HashSet<>(queryParameters.keySet());
-        if (exchange.getRequestMethod().equals(Methods.POST)) {
-            final FormData parsedFormData = parseFormData();
-            if (parsedFormData != null) {
-                Iterator<String> it = parsedFormData.iterator();
-                while (it.hasNext()) {
-                    String name = it.next();
-                    for(FormData.FormValue param : parsedFormData.get(name)) {
-                        if(!param.isFile()) {
-                            parameterNames.add(name);
-                            break;
-                        }
+        final FormData parsedFormData = parseFormData();
+        if (parsedFormData != null) {
+            Iterator<String> it = parsedFormData.iterator();
+            while (it.hasNext()) {
+                String name = it.next();
+                for(FormData.FormValue param : parsedFormData.get(name)) {
+                    if(!param.isFileItem()) {
+                        parameterNames.add(name);
+                        break;
                     }
                 }
             }
@@ -712,15 +759,13 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
                 ret.add(param);
             }
         }
-        if (exchange.getRequestMethod().equals(Methods.POST)) {
-            final FormData parsedFormData = parseFormData();
-            if (parsedFormData != null) {
-                Deque<FormData.FormValue> res = parsedFormData.get(name);
-                if (res != null) {
-                    for (FormData.FormValue value : res) {
-                        if(!value.isFile()) {
-                            ret.add(value.getValue());
-                        }
+        final FormData parsedFormData = parseFormData();
+        if (parsedFormData != null) {
+            Deque<FormData.FormValue> res = parsedFormData.get(name);
+            if (res != null) {
+                for (FormData.FormValue value : res) {
+                    if(!value.isFileItem()) {
+                        ret.add(value.getValue());
                     }
                 }
             }
@@ -740,30 +785,28 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         for (Map.Entry<String, Deque<String>> entry : queryParameters.entrySet()) {
             arrayMap.put(entry.getKey(), new ArrayList<>(entry.getValue()));
         }
-        if (exchange.getRequestMethod().equals(Methods.POST)) {
 
-            final FormData parsedFormData = parseFormData();
-            if (parsedFormData != null) {
-                Iterator<String> it = parsedFormData.iterator();
-                while (it.hasNext()) {
-                    final String name = it.next();
-                    Deque<FormData.FormValue> val = parsedFormData.get(name);
-                    if (arrayMap.containsKey(name)) {
-                        ArrayList<String> existing = arrayMap.get(name);
-                        for (final FormData.FormValue v : val) {
-                            if(!v.isFile()) {
-                                existing.add(v.getValue());
-                            }
+        final FormData parsedFormData = parseFormData();
+        if (parsedFormData != null) {
+            Iterator<String> it = parsedFormData.iterator();
+            while (it.hasNext()) {
+                final String name = it.next();
+                Deque<FormData.FormValue> val = parsedFormData.get(name);
+                if (arrayMap.containsKey(name)) {
+                    ArrayList<String> existing = arrayMap.get(name);
+                    for (final FormData.FormValue v : val) {
+                        if(!v.isFileItem()) {
+                            existing.add(v.getValue());
                         }
-                    } else {
-                        final ArrayList<String> values = new ArrayList<>();
-                        for (final FormData.FormValue v : val) {
-                            if(!v.isFile()) {
-                                values.add(v.getValue());
-                            }
-                        }
-                        arrayMap.put(name, values);
                     }
+                } else {
+                    final ArrayList<String> values = new ArrayList<>();
+                    for (final FormData.FormValue v : val) {
+                        if(!v.isFileItem()) {
+                            values.add(v.getValue());
+                        }
+                    }
+                    arrayMap.put(name, values);
                 }
             }
         }
@@ -827,21 +870,22 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
             if (servletInputStream != null) {
                 throw UndertowServletMessages.MESSAGES.getInputStreamAlreadyCalled();
             }
-            Charset charSet = servletContext.getDeployment().getDefaultCharset();
-            if (characterEncoding != null) {
-                charSet = characterEncoding;
+            Charset charSet = null;
+            if (this.characterEncoding != null) {
+                charSet = this.characterEncoding;
             } else {
-                String c = getCharacterEncodingFromHeader();
+                final String c = getCharacterEncoding();
                 if (c != null) {
                     try {
                         charSet = Charset.forName(c);
                     } catch (UnsupportedCharsetException e) {
-                        throw new UnsupportedEncodingException();
+                        throw new UnsupportedEncodingException(e.getMessage());
                     }
                 }
             }
 
-            reader = new BufferedReader(new InputStreamReader(exchange.getInputStream(), charSet));
+            reader = new BufferedReader(charSet == null ? new InputStreamReader(exchange.getInputStream(), StandardCharsets.ISO_8859_1)
+                    : new InputStreamReader(exchange.getInputStream(), charSet));
         }
         readStarted = true;
         return reader;
@@ -944,18 +988,30 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         return exchange.getSourceAddress().getPort();
     }
 
+    /**
+     * String java.net.InetAddress.getHostName()
+     * Gets the host name for this IP address.
+     * If this InetAddress was created with a host name, this host name will be remembered and returned; otherwise, a reverse name lookup will be performed and the result will be returned based on the system configured name lookup service. If a lookup of the name service is required, call getCanonicalHostName.
+     * If there is a security manager, its checkConnect method is first called with the hostname and -1 as its arguments to see if the operation is allowed. If the operation is not allowed, it will return the textual representation of the IP address.
+     * @see InetAddres#getHostName
+     */
     @Override
     public String getLocalName() {
-        return exchange.getDestinationAddress().getHostString();
+        return exchange.getDestinationAddress().getHostName();
     }
 
     @Override
     public String getLocalAddr() {
-        InetSocketAddress address = exchange.getDestinationAddress();
-         if (address != null) {
-            return address.getAddress().getHostAddress();
+        InetSocketAddress destinationAddress = exchange.getDestinationAddress();
+        if (destinationAddress == null) {
+            return "";
         }
-        return null;
+        InetAddress address = destinationAddress.getAddress();
+        if (address == null) {
+            //this is unresolved, so we just return the host name
+            return destinationAddress.getHostString();
+        }
+        return address.getHostAddress();
     }
 
     @Override
@@ -1008,7 +1064,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
 
     @Override
     public boolean isAsyncStarted() {
-        return asyncStarted || asyncCancelled;
+        return asyncStarted;
     }
 
     @Override
@@ -1016,10 +1072,6 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         return exchange.getAttachment(ServletRequestContext.ATTACHMENT_KEY).isAsyncSupported();
     }
 
-    void setAsyncCancelled(boolean asyncCancelled) {
-        this.asyncCancelled = asyncCancelled;
-    }
-
     @Override
     public AsyncContextImpl getAsyncContext() {
         if (!isAsyncStarted()) {
@@ -1134,9 +1186,32 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         return "HttpServletRequestImpl [ " + getMethod() + ' ' + getRequestURI() + " ]";
     }
 
-    public void clearAttributes() {
-        if(attributes != null) {
-            this.attributes.clear();
+    @Override
+    public PushBuilder newPushBuilder() {
+        if(exchange.getConnection().isPushSupported()) {
+            return new PushBuilderImpl(this);
         }
+        return null;
+    }
+
+    @Override
+    public Map<String, String> getTrailerFields() {
+        HeaderMap trailers = exchange.getAttachment(HttpAttachments.REQUEST_TRAILERS);
+        if(trailers == null) {
+            return Collections.emptyMap();
+        }
+        Map<String, String> ret = new HashMap<>();
+        for(HeaderValues entry : trailers) {
+            ret.put(entry.getHeaderName().toString().toLowerCase(Locale.ENGLISH), entry.getFirst());
+        }
+        return ret;
+    }
+
+    @Override
+    public boolean isTrailerFieldsReady() {
+        if(exchange.isRequestComplete()) {
+            return true;
+        }
+        return !exchange.getConnection().isRequestTrailerFieldsSupported();
     }
 }
