@@ -154,7 +154,6 @@ public class AnnotatedEndpointFactory {
                                 throw JsrWebSocketMessages.MESSAGES.moreThanOneAnnotation(OnMessage.class);
                             }
                             textMessage = new BoundMethod(method, param, true, maxMessageSize, new BoundSingleParameter(method, Session.class, true),
-                                    new BoundSingleParameter(method, boolean.class, true),
                                     new BoundSingleParameter(i, param),
                                     createBoundPathParameters(method, paths, endpointClass));
                             messageHandled = true;
@@ -164,7 +163,6 @@ public class AnnotatedEndpointFactory {
                                 throw JsrWebSocketMessages.MESSAGES.moreThanOneAnnotation(OnMessage.class);
                             }
                             binaryMessage = new BoundMethod(method, param, true, maxMessageSize, new BoundSingleParameter(method, Session.class, true),
-                                    new BoundSingleParameter(method, boolean.class, true),
                                     new BoundSingleParameter(i, param),
                                     createBoundPathParameters(method, paths, endpointClass));
                             messageHandled = true;
@@ -197,7 +195,6 @@ public class AnnotatedEndpointFactory {
                             }
                             binaryMessage = new BoundMethod(method, InputStream.class, false,
                                     maxMessageSize, new BoundSingleParameter(method, Session.class, true),
-                                    new BoundSingleParameter(method, boolean.class, true),
                                     new BoundSingleParameter(i, InputStream.class),
                                     createBoundPathParameters(method, paths, endpointClass));
                             messageHandled = true;
@@ -220,7 +217,6 @@ public class AnnotatedEndpointFactory {
                             }
                             textMessage = new BoundMethod(method, Reader.class, false,
                                     maxMessageSize, new BoundSingleParameter(method, Session.class, true),
-                                    new BoundSingleParameter(method, boolean.class, true),
                                     new BoundSingleParameter(i, Reader.class),
                                     createBoundPathParameters(method, paths, endpointClass));
                             messageHandled = true;
@@ -265,8 +261,8 @@ public class AnnotatedEndpointFactory {
 
 
     private static String[] pathParams(final Method method) {
-        String[] params = new String[method.getParameterTypes().length];
-        for (int i = 0; i < method.getParameterTypes().length; ++i) {
+        String[] params = new String[method.getParameterCount()];
+        for (int i = 0; i < method.getParameterCount(); ++i) {
             PathParam param = getPathParam(method, i);
             if (param != null) {
                 params[i] = param.value();
@@ -317,7 +313,7 @@ public class AnnotatedEndpointFactory {
         BoundSingleParameter(final Method method, final Class<?> type, final boolean optional) {
             this.type = type;
             int pos = -1;
-            for (int i = 0; i < method.getParameterTypes().length; ++i) {
+            for (int i = 0; i < method.getParameterCount(); ++i) {
                 boolean pathParam = false;
                 for (Annotation annotation : method.getParameterAnnotations()[i]) {
                     if (annotation.annotationType().equals(PathParam.class)) {
